import 'dart:async';

import 'package:flutter/material.dart';
import 'package:provider/provider.dart';
import 'package:rxdart/subjects.dart';

class Query<T> {
  final String key;
  final List<Function> observers = [];
  final BehaviorSubject<AsyncSnapshot> _resultStream =
      BehaviorSubject.seeded(AsyncSnapshot.nothing(), sync: true);

  QueryConfig config;

  Query(this.key, this.config);

  Future<T> fetch() {
    Future<T> queryFn;
    if (!_resultStream.value.hasData) {
      _resultStream.add(AsyncSnapshot.waiting());
    }

    if (config.isInfinite) {
      final cursor = config.getFetchMore(_resultStream.value.data);
      if (cursor != false) {
        queryFn = config.fn(cursor);
      }
    } else {
      queryFn = config.fn();
    }

    return queryFn.then((data) {
      final current = _resultStream.value.data;
      if (config.isInfinite) {
        (current as List ?? []).addAll((data as List));
      }

      _resultStream.add(AsyncSnapshot.withData(ConnectionState.active, data));
      return data;
    }).catchError((err) {
      _resultStream.add(AsyncSnapshot.withError(ConnectionState.done, err));
    });
  }

  StreamSubscription<AsyncSnapshot> watch(Function observer) {
    observers.add(observer);
    return _resultStream.listen((snapshot) {
      observer(snapshot);
    });
  }
}

class QueryConfig {
  final isInfinite;
  final Function fn;
  final Function getFetchMore;

  QueryConfig({
    @required this.fn,
    this.getFetchMore,
    this.isInfinite = false,
  });
}

class QueryCache {
  Map<String, Query> _queryMap = {};

  Query getQueryByKey(String key) {
    return _queryMap[key];
  }

  Query registerQuery(String key, QueryConfig config) {
    var query = getQueryByKey(key);
    if (query == null) {
      query = Query(key, config);
      _queryMap[key] = query;
      invalidateQuery(key);
    }
    return query;
  }

  StreamSubscription watchQuery(String key, Function observer) {
    var query = getQueryByKey(key);
    if (query != null) {
      return query.watch(observer);
    }
    return null;
  }

  void unwatchQuery(String key, Function observer) {
    var query = getQueryByKey(key);
    if (query != null) {
      query.observers.remove(observer);
    }
  }

  void invalidateQuery(String key) {
    final query = getQueryByKey(key);
    if (query != null) {
      query.fetch();
    }
  }
}

class QueryCacheProvider extends StatefulWidget {
  final Widget child;
  QueryCacheProvider({this.child});

  @override
  State<StatefulWidget> createState() => QueryCacheProviderState();
}

class QueryCacheProviderState extends State<QueryCacheProvider> {
  QueryCache _queryCache;

  @override
  void initState() {
    _queryCache = QueryCache();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Provider.value(
      value: _queryCache,
      child: widget.child,
    );
  }
}
